#include "parse_event_handler.h"
#include "expat_module.h"
#include "domlette.h"
#include "xptr.h"
#include "util.h"

/** forwards ***********************************************************/

static void processingInstructionStateHandler(void *userState, void *params);
static void characterDataStateHandler(void *userState, void *params);
static void startNsScopeStateHandler(void *userState, void *params);
static void endNsScopeStateHandler(void *userState, void *params);
static void startElementStateHandler(void *userState, void *params);
static void endElementStateHandler(void *userState, void *params);
static void commentStateHandler(void *userState, void *params);

static void completeText(ParserState *state);
static void xinclude(ParserState *state);

#ifdef DEBUG_PARSER
static void dumpParserState(ParserState *state);
#endif

/** handlers ***********************************************************/

static void startNsScopeStateHandler(void *userState, void *params)
{
  ParserState *state = (ParserState *)userState;
  PyObject *xml = PyUnicode_DecodeASCII("xml", 3, NULL);
  /*NsMapping *nsm = NULL;*/
  
  /*
    This is a kludge to get around what appears to be a bug in expat: it
    welds "xml=http://www.w3.org/XML/1998/namespace" to the beggining
    of a context for an ext entity parser, which then proceeds to
    read this (why would it be parsing from the context???) and
    send a startNsScope request.  Even more oddly, it then omits to send
    an endNsScope request, causing memory management problems
  */

  if (!PyObject_Compare(state->expat_prefix, xml)) {
#ifdef DEBUG_PARSER
    fprintf(stderr, "NSDecl ignored\n");
#endif
    Py_DECREF(xml);
    return;
  }
  Py_DECREF(xml);

#ifdef DEBUG_PARSER
  fprintf(stderr, "NSDecl started\n");
  fprintf(stderr, "URI:\n");
  _PyObject_Dump(state->expat_uri);
  fprintf(stderr, "Prefix:\n");
  _PyObject_Dump(state->expat_prefix);
#endif

  /* We don't need to store the namespace/prefix mapping, as the prefix
   * is now part of the name from expat.
  nsm = PyMem_New(NsMapping, 1);
  if (nsm == NULL) {
    signalError(state, "Out Of Memory");
    return; 
  }
  nsm->uri = state->expat_uri;
  nsm->prefix = state->expat_prefix;

  list_ins_next(state->namespaces, NULL, nsm);
  Py_INCREF(nsm->uri);
  Py_INCREF(nsm->prefix);
  */

  PyDict_SetItem(state->new_nss, state->expat_prefix, state->expat_uri);
  return;
}


static void endNsScopeStateHandler(void *userState, void *params)
{
  ParserState *state = (ParserState *)userState;
  PyObject *xml = PyUnicode_DecodeASCII("xml", 3, NULL);
  /*NsMapping *nsm;*/

  /* See comment in startNsScope */
  if (!PyObject_Compare(state->expat_prefix, xml)) {
#ifdef DEBUG_PARSER
    fprintf(stderr, "NSDecl ignored\n");
#endif
    Py_DECREF(xml);
    return;
  }
  Py_DECREF(xml);

#ifdef DEBUG_PARSER
  fprintf(stderr, "NSDecl Ended\n");
  fprintf(stderr, "Prefix:\n");
  _PyObject_Dump(state->expat_prefix);
#endif
  /*
  list_find_remove(state->namespaces, 
                   (void *)state->expat_prefix, 
                   (void *)&nsm,
                   matchPrefix);
  Py_DECREF(nsm->uri);
  Py_DECREF(nsm->prefix);
  PyMem_Del(nsm); 
  */
  return;
}


static void startElementStateHandler(void *userState, void *params)
{
  ParserState *state = (ParserState *)userState;
  PyElementObject *elem = NULL;
  UniversalName *un;
  int i;
  int *new_preserve_state;
  static PyObject *processIncludes, *xi_namespace, *xi_include;
  static PyObject *asterisk, *stripElements;
  PyObject *key, *value;
  PyObject *strip_list;
  PyObject *xml_base;

#ifdef DEBUG_PARSER
  fprintf(stderr, "Start element\n");
  fprintf(stderr, "name:\n");
  _PyObject_Dump(state->expat_name);
#endif

  /* for getattr (needs to be PyString) */
  if (processIncludes == NULL) {
    processIncludes = PyString_InternFromString("processIncludes");
    if (processIncludes == NULL) {
      signalError(state, "Out of Memory");
      return;
    }
  }

  /* for getattr (needs to be PyString) */
  if (stripElements == NULL) {
    stripElements = PyString_InternFromString("stripElements");
    if (stripElements == NULL) {
      signalError(state, "Out of Memory");
      return;
    }
  }

  if (xi_namespace == NULL) {
    xi_namespace = PyUnicode_DecodeASCII(XINCLUDE_NAMESPACE, strlen(XINCLUDE_NAMESPACE), NULL);
    if (xi_namespace == NULL) {
      signalError(state, "Out of Memory");
      return;
    }
  }

  if (xi_include == NULL) {
    xi_include = PyUnicode_DecodeASCII("include", 7, NULL);
    if (xi_include == NULL) {
      signalError(state, "Out of Memory");
      return;
    }
  }

  if (asterisk == NULL) {
    asterisk = PyUnicode_DecodeASCII("*", 1, NULL);
    if (asterisk == NULL) {
      signalError(state, "Out of Memory");
      return;
    }
  }

  /* Call Complete Text to make sure there are no additional text nodes that need to be created*/
  completeText(state);

  /* Call build universal name to get the current NS. */
  un = buildUniversalName(state, state->expat_name);
  if (!un) {
    /*Should have already transitioned to an error*/
    return;
  }

  /** XInclude processing **********************************************/

  value = PyObject_GetAttr(state->input_source, processIncludes);
  if (PyObject_IsTrue(value) && !PyObject_Compare(un->uri, xi_namespace) && 
      !PyObject_Compare(un->local, xi_include)) {
    Py_DECREF(value);
    destroyUniversalName(un);
    xinclude(state);
    return;
  }
  Py_DECREF(value);
  /*
  if (state->xinclude_content_depth) {
    state->xinclude_content_depth += 1;
    destroyUniversalName(un);
    return;
  }
  */
  /* Now create the element */
  elem = Document_CreateElementNS(state->ownerDoc, 
                                  un->uri, un->prefix, un->local, 
                                  state->docIx);
  destroyUniversalName(un);

  if (elem == NULL) {
    signalError(state,"Out of Memory");
    return;
  }

  /* Check for preserve state */
  if ((new_preserve_state = (int *)malloc(sizeof(int))) == NULL) {
    Py_DECREF(elem);
    signalError(state,"Out of Memory");
    return;
  }

  *new_preserve_state = *((int *)_stack_peek(state->preserve_state));
  xml_base = (PyObject *)_stack_peek(state->xml_base_stack);

  /** XSLT strip-elements **********************************************/

  /* stripElements is a list of tuples */
  strip_list = PyObject_GetAttr(state->input_source, stripElements);
  for (i=0; i < PyList_GET_SIZE(strip_list); i++) { 
    PyObject *strip_item = PyList_GET_ITEM(strip_list, i);
    PyObject *strip_uri = PyTuple_GET_ITEM(strip_item, 0);
    PyObject *strip_local = PyTuple_GET_ITEM(strip_item, 1);
    PyObject *strip_flag = PyTuple_GET_ITEM(strip_item, 2);

    /* Each strip element specifies a NS and a local name,
       The localname can be a name or *
       If the localname is * then the NS could be None.
       ns:local is a complete match
       ns:* matches any element in the given namespace
       * matches any element.
   
       NOTE:  There are precedence rules to stripping as (according to XSLT)
       the matches should be treated as node tests.  The precedence is based
       on level of specificity.  This code assumes the list of white space
       rules has already been sorted by precedence.
    */
     
    if (/* if localName == '*' and no namespace or the namespace matches, or */
        (!PyObject_Compare(strip_local, asterisk) && 
         (strip_uri == Py_None || 
          !PyObject_Compare(strip_uri, elem->namespaceURI))) ||
        /* the localName and namespaceURI match */
        (!PyObject_Compare(strip_local, elem->localName) && 
         !PyObject_Compare(strip_uri, elem->namespaceURI))) {
      *new_preserve_state = !PyInt_AS_LONG(strip_flag);
      break;
    }
  }
  Py_DECREF(strip_list);
  
  /** namespaces *******************************************************/

  /* new_nss is a dictionary where key is the prefix and value is the uri */
  i = 0;
  while (PyDict_Next(state->new_nss, &i, &key, &value)) {
    PyAttrObject *nsattr;
    PyObject *prefix, *localName;

    Py_INCREF(key);
    if (key != Py_None) {
      prefix = PyUnicode_DecodeASCII("xmlns", 5, NULL);
      localName = key;
    } else {
      prefix = key;
      localName = PyUnicode_DecodeASCII("xmlns", 5, NULL);
    }

    nsattr = Element_SetAttributeNS(elem, g_xmlnsNamespace, localName, prefix,
                                    value);
    Py_DECREF(nsattr);
    Py_DECREF(localName);
    Py_DECREF(prefix);
  }
  /* make sure children don't set these namespaces */
  PyDict_Clear(state->new_nss);

  /** attributes *******************************************************/

  /* expat_atts is a dictionary where key is the expanded name and value 
     is the attribute value.
  */
  i = 0;
  while (PyDict_Next(state->expat_atts, &i, &key, &value)) {
    PyAttrObject *attr;
    un = buildUniversalName(state, key);
    if (!un) {
      Py_DECREF(elem);
      return;
    }

    attr = Element_SetAttributeNS(elem, un->uri, un->local, un->prefix, value);
    destroyUniversalName(un);
    
    /* check for xml:space and xml:base */
    if (!PyObject_Compare(attr->namespaceURI, g_xmlNamespace)) {
      PyObject *base = PyUnicode_DecodeASCII("base", 4, NULL);
      PyObject *space = PyUnicode_DecodeASCII("space", 5, NULL);

      if (!PyObject_Compare(attr->localName, base)) {
        /* use a borrowed value to simplify cleanup */
        xml_base = value;
      } else if (!PyObject_Compare(attr->localName, space)) {
        PyObject *space_preserve = PyUnicode_DecodeASCII("preserve", 8, NULL);
        PyObject *space_default = PyUnicode_DecodeASCII("default", 7, NULL);
        if (!PyObject_Compare(value, space_preserve))
          *new_preserve_state = 1;
        else if (!PyObject_Compare(value, space_default))
          *new_preserve_state = 0;;
        Py_DECREF(space_preserve);
        Py_DECREF(space_default);
      }
      Py_DECREF(base);
      Py_DECREF(space);
    }
    Py_DECREF(attr);
  }

  /* Set XML base */
  Py_DECREF(elem->xmlBase);
  Py_INCREF(xml_base);
  elem->xmlBase = xml_base;

  /* save states on their respective stacks */
  _stack_push(state->preserve_state, new_preserve_state);
  _stack_push(state->xml_base_stack, xml_base);
  _stack_push(state->node_stack, elem);
  _stack_push(state->elem_depth_event_stack, state->curr_elem_depth_event);
  return;
}


static void endElementStateHandler(void *userState, void *params)
{
  ParserState *state = (ParserState *)userState;
  int *top_ps;
  int *event;
  void *ptr;
  PyNodeObject *new_elem;
  PyNodeObject *top_node;

#ifdef DEBUG_PARSER
  fprintf(stderr, "End element\n");
  fprintf(stderr, "name:\n");
  _PyObject_Dump(state->expat_name);
#endif

  if (state->xinclude_content_depth){
    state->xinclude_content_depth--;
    return;
  }

  /*Complete Any outstanding text nodes*/
  completeText(state);

  /*Pop the last element off of the stack*/
  _stack_pop(state->node_stack, (void *)&new_elem);

  /*Pop off the preserve state*/
  if (!_stack_pop(state->preserve_state, &ptr)){
    top_ps = (int *)ptr;
    free(top_ps);
  }

  /*Pop off the xml base state*/
  if (!_stack_pop(state->xml_base_stack, &ptr)){
    /*Py_DECREF((PyObject *)ptr);*/
  }

  /*Append us to the tree*/
  top_node = (PyNodeObject *)_stack_peek(state->node_stack);
  Node_AppendChild(top_node, new_elem);
  /*Append child adds a REF*/
  Py_DECREF(new_elem);

  /*Make any special state transitions associated with the end of this element */
  if (!_stack_pop(state->elem_depth_event_stack, &ptr)){
    event = (int *)ptr;
    if (*event) transit(state, *event);
    free(event);
  }
}


static void characterDataStateHandler(void *userState, void *params)
{
  ParserState *state = (ParserState *)userState;
  PyObject *temp;

#ifdef DEBUG_PARSER
  fprintf(stderr, "New character data: ");
  fprintf(stderr,"Original: \n");
  _PyObject_Dump(state->curr_text);
  fprintf(stderr,"Added: \n");
  _PyObject_Dump(state->expat_charbuf);
#endif

  if (state->curr_text) {
    /* existing text */
    temp = PySequence_Concat(state->curr_text, (PyObject *)state->expat_charbuf);
    Py_DECREF(state->curr_text);
    state->curr_text = temp;
  } else {
    /* first text */
    state->curr_text = (PyObject *)state->expat_charbuf;
    Py_INCREF(state->curr_text);
  }

#ifdef DEBUG_PARSER
  fprintf(stderr, "New: ");
  _PyObject_Dump(state->curr_text);
#endif
}


static void processingInstructionStateHandler(void *userState, void *params)
{
  ParserState *state = (ParserState *)userState;
  PyNodeObject *new_pi;
  PyNodeObject *top_node;

  completeText(state);

  new_pi =(PyNodeObject *)Document_CreateProcessingInstruction(state->ownerDoc,
							       state->expat_target,
							       state->expat_data,
							       state->docIx);
  top_node = (PyNodeObject *)_stack_peek(state->node_stack);
  Node_AppendChild(top_node, new_pi);

  /*Node Append Child added a ref*/
  Py_DECREF(new_pi);

}


static void commentStateHandler(void *userState, void *params)
{
  ParserState *state = (ParserState *)userState;
  PyNodeObject *new_comment;
  PyNodeObject *top_node;

  completeText(state);

  new_comment = (PyNodeObject *)Document_CreateComment(state->ownerDoc,
                                                       state->expat_data,
                                                       state->docIx);
  top_node = (PyNodeObject *)_stack_peek(state->node_stack);
  Node_AppendChild(top_node, new_comment);
  /*Append Child adds a ref*/
  Py_DECREF(new_comment);
}


/* local support routines */

#define IS_XMLSPACE(c) (((c) == 0x09) || \
			((c) == 0x0A) || \
			((c) == 0x0D) || \
			((c) == 0x20))

static void completeText(ParserState *state)
{
  PyNodeObject *new_text;
  PyNodeObject *top_node;
  int preserve_state;
  int i;

  if (state->curr_text) {
    preserve_state = *((int *)_stack_peek(state->preserve_state));

    /* Always preserve the text node if it contains non XML space chars */
    for (i = 0; i < PyUnicode_GET_SIZE(state->curr_text); i++) {
      Py_UNICODE ch = PyUnicode_AS_UNICODE(state->curr_text)[i];
      if (!IS_XMLSPACE(ch)) {
	preserve_state = 1;
	break;
      }
    }

    if (preserve_state) {
      /* New reference */
      new_text = (PyNodeObject *)Document_CreateTextNode(state->ownerDoc,
							 state->curr_text,
							 state->docIx);

      top_node = (PyNodeObject *)_stack_peek(state->node_stack);

      /* Adds reference */
      Node_AppendChild(top_node, new_text);

      Py_DECREF(new_text);
    }
    Py_DECREF(state->curr_text);
    state->curr_text = NULL;
  }
}


static void xinclude(ParserState *state)
{
  /*Create a sub processor to handle the include*/

  char buf[BUFSIZ];
  XML_Parser xi_parser = XML_ParserCreateNS(NULL, EXPAT_NSSEP);
  int i, xi_parse_text=0, xi_start_state=PARSE_STREAM_STATE;
  PyObject *new_input_source, *data, *hint, *stream, *current_input_source;
  size_t len;
  int done;
  PyObject *attr_name, *attr_value, *href = NULL;
  PyObject *save_name, *save_atts;
  PyObject *modules, *cdom_mod, *xptr_spec;

  PyObject *href_str = PyUnicode_DecodeASCII("href", 4, NULL);
  PyObject *encoding_str = PyUnicode_DecodeASCII("encoding", 8, NULL);
  PyObject *parse_str = PyUnicode_DecodeASCII("parse", 5, NULL);
  PyObject *text_str = PyUnicode_DecodeASCII("text", 4, NULL);
  PyObject *xml_str = PyUnicode_DecodeASCII("xml", 3, NULL);
  
  /* encodings MUST be ascii per the XML spec so OK to use a PyString */
  PyObject *encoding = PyString_FromString("UTF-8");

  href = NULL;
  i = 0;
  while (PyDict_Next(state->expat_atts, &i, &attr_name, &attr_value)) {
    /* because these attributes have no namespace, it is ok to compare the
       name from expat.
    */
    if (PyObject_Compare(attr_name, href_str) == 0) {
      href = attr_value;
      Py_INCREF(href);
    } else if (PyObject_Compare(attr_name, encoding_str) == 0) {
      /* delete the previous encoding */
      Py_DECREF(encoding);
      /* encodings MUST be ascii per the XML spec */
      encoding = PyUnicode_AsASCIIString(attr_value);
    } else if (PyObject_Compare(attr_name, parse_str) == 0) {
      if (PyObject_Compare(attr_value, text_str) == 0)
        xi_parse_text = 1;
      else if (PyObject_Compare(attr_value, xml_str) != 0) {
        Py_XDECREF(href);
        Py_DECREF(encoding);
        Py_DECREF(href_str);
        Py_DECREF(encoding_str);
        Py_DECREF(parse_str);
        Py_DECREF(text_str);
        Py_DECREF(xml_str);
        signalError(state, "Invalid XInclude parse attribute value");
        return;
      }
    }
  }
  Py_DECREF(href_str);
  Py_DECREF(encoding_str);
  Py_DECREF(parse_str);
  Py_DECREF(text_str);
  Py_DECREF(xml_str);

  if (href == NULL) {
    /* Early exit if href attr is missing*/
    Py_DECREF(encoding);
    signalError(state, "Missing XInclude href attribute");
    return;
  }

  /* Invoke resolve on the input_source to get the next input source*/
  hint = PyString_FromString("XINCLUDE");

  new_input_source = PyObject_CallMethod(state->input_source, "resolve",
                                         "OOOi", href, Py_None, hint, 0);
  Py_DECREF(hint);

  if (new_input_source == NULL) {
    /* FIXME: Propagate the current error object */
    Py_DECREF(encoding);
    signalError(state, "Error in the entity resolver");
    return;
  }

  /* Get any XPointer fragment */
  modules = PyImport_GetModuleDict(); /* borrowed */
  cdom_mod = PyDict_GetItemString(modules, "Ft.Xml.cDomlette"); /* borrowed */
  if (!cdom_mod) {
    cdom_mod = PyImport_ImportModule("Ft.Xml.cDomlette");
    if (!cdom_mod) 
      signalError(state, "Error in importing cDomlette");
    /* Make it a borrowed reference, a copy exists in sys.modules */
    Py_DECREF(cdom_mod);
  }
  
  xptr_spec = PyObject_CallMethod(cdom_mod, "ProcessFragment", "O", href);
  /* We shall use xptr_spec as the key to holding all the refs of the state specs */
  if (!xptr_spec){
    signalError(state, "Error in Xpointer expresison for XInclude");
  }
  if (xptr_spec != Py_None){
    xi_start_state = handle_xpointer(state, xptr_spec);
    if (!xi_start_state)
      signalError(state, "Error preparing state machine for XInclude");
  }
  Py_DECREF(xptr_spec);
  Py_DECREF(href);

#ifdef DEBUG_PARSER
  fprintf(stderr, "Begin XInclude parsing.\n");
#endif

  /* Get the stream from the new input source*/
  stream = PyObject_GetAttrString(new_input_source, "stream");
  Py_INCREF(stream);

  transit(state, PARSE_RESUME_EVENT);
  if (xi_parse_text) {
    do {
      len = readFromObject(stream, buf, BUFSIZ);
      if (PyErr_Occurred()) {
        Py_DECREF(new_input_source);
        Py_DECREF(encoding);
        Py_DECREF(stream);
        return;
      }
      done = len < sizeof(buf);
      /* A bit convoluted, but we must consider matters such as ws stripping,
         so better to handle this using the normal parse handling flow*/
      data = PyUnicode_Decode(buf, len, PyString_AS_STRING(encoding), NULL);
      if (data) {
	state->expat_charbuf = data;
	transit(state, CHARACTER_DATA_EVENT);
      }
      if (!data || PyErr_Occurred() || state->core.current == ERROR_STATE) {
        Py_XDECREF(data);
        Py_DECREF(encoding);
        Py_DECREF(stream);
        Py_DECREF(new_input_source);
        signalError(state, "");
        return;
      }
      transit(state, PARSE_RESUME_EVENT);
      Py_DECREF(data);
    } while (!done);
  } else {
    /*Save a copy of the current input source*/
    current_input_source = state->input_source;
    state->input_source = new_input_source;

    initParser(&xi_parser, state);

    /*Reset the state machine, save name and attrs*/
    save_name = state->expat_name;
    save_atts = state->expat_atts;
    resetState(state, xi_start_state);

    do {
      len = readFromObject(stream, buf, BUFSIZ);
      if (PyErr_Occurred()) {
        Py_DECREF(new_input_source);
        Py_DECREF(encoding);
        Py_DECREF(stream);
        return;
      }
      done = len < sizeof(buf);
      if (!XML_Parse(xi_parser, buf, len, done)) {
        PyObject *uri = PyObject_GetAttrString(new_input_source, "uri");
        PyObject *repr = PyObject_Repr(uri);
        Py_DECREF(uri);
	/* Exception: Well-formedness error*/
        /*PyErr_Format(g_readerErrorObject, ...);*/
        PyErr_Format(PyExc_SyntaxError, "%s\t%d\t%d\t%s",
                     PyString_AsString(repr),
                     XML_GetCurrentLineNumber(xi_parser),
                     XML_GetCurrentColumnNumber(xi_parser),
                     XML_ErrorString(XML_GetErrorCode(xi_parser)));
        Py_DECREF(repr);
        /*Set the state back*/
	state->input_source = current_input_source;
        state->expat_name = save_name;
        state->expat_atts = save_atts;

        Py_DECREF(new_input_source);
        Py_DECREF(encoding);
        Py_DECREF(stream);
        signalError(state, "Parse error in XInclude");
        return;
      }
      if (PyErr_Occurred() || state->core.current == ERROR_STATE) {
        Py_DECREF(new_input_source);
        Py_DECREF(encoding);
        Py_DECREF(stream);
        return;
      }
    } while (!done);

    XML_ParserFree(xi_parser);

    /*All done with the new input source*/
    Py_DECREF(new_input_source);  

    /*Set the state back*/
    state->input_source = current_input_source;
    state->expat_name = save_name;
    state->expat_atts = save_atts;
    resetState(state, START_ELEMENT_CALLBACK);
  }

  Py_DECREF(stream);       /* refcnt -> 0 */

#ifdef DEBUG_PARSER
  fprintf(stderr, "End XInclude parsing.\n");
#endif

  Py_DECREF(encoding);
  state->xinclude_content_depth++;
  return;
}


/* debugging aids */
#ifdef DEBUG_PARSER
static void dumpParserState(ParserState *state)
{
  fprintf(stderr, "Current Text: ");
  _PyObject_Dump(state->curr_text);
  fprintf(stderr, "\n");
}
#endif

/* State table set-up */


int initializeStateTableForParseEvents(ParserState *state)
{
  StateTableEntry *ste;

  /* Add new states and handlers */
  if ((ste = newStateTableEntry(state, START_ELEMENT_CALLBACK)) == NULL)
    return 0;
  ste->handler = startElementStateHandler;
  if ((ste = newStateTableEntry(state, END_ELEMENT_CALLBACK)) == NULL)
    return 0;
  ste->handler = endElementStateHandler;
  if ((ste = newStateTableEntry(state, START_NS_SCOPE_CALLBACK)) == NULL)
    return 0;
  ste->handler = startNsScopeStateHandler;
  if ((ste = newStateTableEntry(state, END_NS_SCOPE_CALLBACK)) == NULL)
    return 0;
  ste->handler = endNsScopeStateHandler;
  if ((ste = newStateTableEntry(state, CHARACTER_DATA_CALLBACK)) == NULL)
    return 0;
  ste->handler = characterDataStateHandler;
  if ((ste = newStateTableEntry(state, PI_CALLBACK)) == NULL)
    return 0;
  ste->handler = processingInstructionStateHandler;
  if ((ste = newStateTableEntry(state, COMMENT_CALLBACK)) == NULL)
    return 0;
  ste->handler = commentStateHandler;

  /* Add new transitions */
  if (!addTransition(state, PARSE_STREAM_STATE, START_ELEMENT_EVENT,
                     START_ELEMENT_CALLBACK))
    return 0;
  if (!addTransition(state, START_ELEMENT_CALLBACK, PARSE_RESUME_EVENT,
                     PARSE_STREAM_STATE))
    return 0;
  if (!addTransition(state, PARSE_STREAM_STATE, END_ELEMENT_EVENT,
                     END_ELEMENT_CALLBACK))
    return 0;
  if (!addTransition(state, END_ELEMENT_CALLBACK, PARSE_RESUME_EVENT,
                     PARSE_STREAM_STATE))
    return 0;

  if (!addTransition(state, PARSE_STREAM_STATE, START_NS_SCOPE_EVENT,
                     START_NS_SCOPE_CALLBACK))
    return 0;
  if (!addTransition(state, START_NS_SCOPE_CALLBACK, PARSE_RESUME_EVENT,
                     PARSE_STREAM_STATE))
    return 0;
  if (!addTransition(state, PARSE_STREAM_STATE, END_NS_SCOPE_EVENT,
                     END_NS_SCOPE_CALLBACK))
    return 0;
  if (!addTransition(state, END_NS_SCOPE_CALLBACK, PARSE_RESUME_EVENT,
                     PARSE_STREAM_STATE))
    return 0;

  if (!addTransition(state, PARSE_STREAM_STATE, CHARACTER_DATA_EVENT,
                     CHARACTER_DATA_CALLBACK))
    return 0;
  if (!addTransition(state, CHARACTER_DATA_CALLBACK, PARSE_RESUME_EVENT,
                     PARSE_STREAM_STATE))
    return 0;

  if (!addTransition(state, PARSE_STREAM_STATE, PI_EVENT,
                     PI_CALLBACK))
    return 0;
  if (!addTransition(state, PI_CALLBACK, PARSE_RESUME_EVENT,
                     PARSE_STREAM_STATE))
    return 0;

  if (!addTransition(state, PARSE_STREAM_STATE, COMMENT_EVENT,
                     COMMENT_CALLBACK))
    return 0;
  if (!addTransition(state, COMMENT_CALLBACK, PARSE_RESUME_EVENT,
                     PARSE_STREAM_STATE))
    return 0;
  return 1;
}


/*
  Create a Universal Name structure from the expat name.
  If the name from the document had a prefix, then expat name is:
    namespace-uri + sep + localName + sep + prefix,
  otherwise if there is a default namespace:
    namespace-uri + sep + localName
  lastly, it could just be:
    localName
  
  If there is a prefix look for the namespace
  In the namespace stack.
  If there is no prefix, set NS to None
 */

UniversalName *buildUniversalName(ParserState *state, PyObject *expanded_name)
{
  Py_UNICODE nssep = EXPAT_NSSEP;
  int len = PyUnicode_GET_SIZE(expanded_name);
  const Py_UNICODE *p = PyUnicode_AS_UNICODE(expanded_name);
  int i, j;
  UniversalName *un;
  /*ListElmt *node = NULL;*/
  
  un = PyMem_New(UniversalName, 1);
  if (un == NULL) {
    signalError(state,"Out of Memory");
    return NULL;
  }
  
  /* scan for beginning of localName */
  for (i = 0; i < len && p[i] != nssep; i++);

  if (i == len) {
    /* just a localName, set namespace and prefix to None. */
    un->uri = Py_None;
    un->local = expanded_name;
    un->prefix = Py_None;
    Py_INCREF(un->uri);
    Py_INCREF(un->local);
    Py_INCREF(un->prefix);
    return un;
  }

  /* found a namespace uri */
  un->uri = PyUnicode_FromUnicode(p, i);
  i++;

  for (j = i; j < len && p[j] != nssep; j++);
  
  un->local = PyUnicode_FromUnicode(p + i, j - i);
  j++;

  if (j < len) {
    /* a prefix is given as well */
    un->prefix = PyUnicode_FromUnicode(p + j, len - j);
  } else {
    un->prefix = Py_None;
    Py_INCREF(Py_None);
  }
#ifdef DEBUG_PARSER
  printf("-- Universal Name ------------------------------------------\n");
  printf("namespace-uri: "); PyObject_Print(un->uri, stdout, Py_PRINT_RAW);
  printf("\nlocal-name: "); PyObject_Print(un->local, stdout, Py_PRINT_RAW);
  printf("\nprefix: "); PyObject_Print(un->prefix, stdout, Py_PRINT_RAW);
  printf("\n");
#endif
  return un;
}

void destroyUniversalName(UniversalName *un)
{
#ifdef DEBUG_PARSER
  fprintf(stderr, "destroyUniversalName\n");
  _PyObject_Dump(un->local);
  _PyObject_Dump(un->uri);
  _PyObject_Dump(un->prefix);
#endif
  Py_XDECREF(un->prefix);
  Py_XDECREF(un->local);
  Py_XDECREF(un->uri);
  PyMem_Del(un);
#ifdef DEBUG_PARSER
  fprintf(stderr, "end destroyUniversalName\n");
#endif
}

