#include "Python.h"
#include "state_machine.h"

void errorHandler(void *userState, void *params);


int initializeStateTable(void *userState) {
  State *state = (State *)userState;
  int i;
  StateTableEntry *es;

  /*Initalize all to None so that we know what to free when we are done*/
  for (i=0;i<MAX_STATE;i++) {
    state->table[i] = NULL;
  }

  /*Add the error state*/
  es = newStateTableEntry(userState, ERROR_STATE);
  if (!es) {
    return 0;
  }
  es->handler = errorHandler;
  return 1;
}

int destroyStateTable(void *userState) {
  State *state = (State *)userState;
  int i;

  /*Initalize all to None so that we know what to free when we are done*/
  for (i=0;i<MAX_STATE;i++) {
    if (state->table[i] != NULL) {
      if (state->table[i]->transitions) {
	free(state->table[i]->transitions);
      }
      free(state->table[i]);
      state->table[i] = NULL;
    }
  }

  return 1;
}


void transit(void *userState, short event)
{
  State *state = (State *)userState;
  short new_state_id  = stateLookupTransition(state, event);
  StateTableEntry *s = state->table[new_state_id];

#ifdef DEBUG_PARSER
  fprintf(stderr, "transition from %i to %i on event %i\n", state->current, new_state_id, event);
#endif
  
  if (s) {
    state->current = new_state_id;
    if (s->handler) {
      s->handler(userState, s->params);
    }
  } else {
    /* switch to error state if there is nowhere else to go */
    state->current = ERROR_STATE;
  }
}


/* Use when all you wish to do is transit to another state */
void plainTransitHandler(void *userState, void *params)
{
  short *next = (short *)params;

  transit(userState, *next);
}


short stateLookupTransition(void *userState, short event)
{
  State *state = (State *)userState;
  StateTableEntry *ste = state->table[state->current];
  short *ptransitions, next=0;

  if (!ste) return ERROR_STATE;
  if (!ste->transitions) return ERROR_STATE;
  /*
    The old logic:
    for (ptransitions=ste->transitions; *ptransitions && !next; ptransitions+=2) {
    New logic below: Instead we want to select the *last* matching transition.
    New logic is slower can speed up by adding new transitions
    to the head of the list and restoring the old logic.
  */
  for (ptransitions=ste->transitions; *ptransitions; ptransitions+=2) {
    if (*ptransitions == event) next=*(ptransitions+1);
  }
  return next ? next : ERROR_STATE;
}


StateTableEntry *newStateTableEntry(void *userState, short id)
{
  State *state = (State *)userState;
  StateTableEntry *ste;

  if ((ste = (StateTableEntry *)calloc(1, sizeof(StateTableEntry))) == NULL) {
    /* Exception: Out of memory */
    return NULL;
  }
  state->table[id] = ste;

  /*Always add a transition from this state to the error state*/
  addTransition(userState, id, ERROR_EVENT, ERROR_STATE);

  return ste;
}


int addTransition(void *userState, short from, short event, short to)
{
  State *state = (State *)userState;
  StateTableEntry *ste = state->table[from];
  short *ptransitions;
  size_t len=0;
  char buffer[256];

#ifdef DEBUG_PARSER
  fprintf(stderr, "add transition from %i to %i on event %i\n", from, to, event);
#endif
  if (!ste) {
    sprintf(buffer, "Unable to set transition from state %d", from);
    PyErr_SetString(PyExc_SystemError, buffer);
    return 0;
  }
  
  ptransitions = ste->transitions;
  if (!ptransitions)
  {
    if ((ptransitions = (short *)calloc(3, sizeof(short))) == NULL) {
      /* Exception: Out of memory */
      PyErr_NoMemory();
      return 0;
    }
    ste->transitions = ptransitions;
  } else {
    while (*ptransitions++) len++;
    ptransitions = ste->transitions;
    if ((ptransitions = (short *)realloc(ptransitions, (len + 3) * sizeof(short))) == NULL) {
      /* Exception: Out of memory */
      PyErr_NoMemory();
      return 0;
    }
    ste->transitions = ptransitions;
    ptransitions += len;    
  }
  *ptransitions = event;
  *(ptransitions+1) = to;
  *(ptransitions+2) = 0;

  return 1;
}


void resetState(void *userState, short new)
{
  State *state = (State *)userState;

  state->current = new;
}

void errorHandler(void *userState, void *params)
{
#ifdef DEBUG_PARSER
  fprintf(stderr,"Pass through Error Handler\n");
#endif
}

